#!/usr/bin/env python3

import re
import os
import sys
import time
import argparse
import threading
import subprocess as sp
from datetime import datetime
from collections import OrderedDict
from inotify_simple import INotify
from inotify_simple import flags as INFlags


"""
rinotify.py: recursive inotify
2019-06-22, v1.0, ed <irc.rizon.net>, MIT-Licensed
https://ocv.me/dev/?rinotify.py

* ignore files/folders with -x REGEX
* run script on changes with -r PATH
  * argv is the list of affected files
* 200 msec debounce with -wd 0.2 (default)
* monitor disk-free with --df

dependencies:
  python3 -m pip install --user inotify_simple
"""


def eprint(*args, end="", file=sys.stderr, **kwargs):
    print(*args, end=end, file=file, **kwargs)


def diskmon(top):
    old = -1024 * 1024 * 1024
    while True:
        time.sleep(0.5)
        st = os.statvfs(top)
        free = st.f_bavail * st.f_frsize
        total = st.f_blocks * st.f_frsize
        used = (st.f_blocks - st.f_bfree) * st.f_frsize

        if abs(used - old) < 1024 * 1024:
            continue

        old = used
        ts = datetime.utcnow().strftime("%Y-%m%d-%H%M%S.%f")
        free = free / (1024 * 1024.0)
        used = used * 100.0 / total
        eprint(f"\033[36m{ts}\033[32m {free:.2f} MB free ({used:.2f}% used)\033[0m\n")


class Sink(object):
    def __init__(self, duration, timeout, cmd, do_out):
        self.duration = duration
        self.timeout = timeout
        self.cmd = cmd
        self.do_stdout = do_out

        self.proc = None
        self.hist = OrderedDict()
        if self.duration > 0:
            self.mtx = threading.Lock()
            threading.Thread(target=self._worker, daemon=True).start()

    def put(self, path, flags):
        item = f"{path} {flags}"
        if self.duration <= 0:
            return self._print({item: 1})

        with self.mtx:
            try:
                self.hist[item] += 1
            except:
                self.hist[item] = 1

    def _print(self, hist):
        ts = datetime.utcnow().strftime("%Y-%m%d-%H%M%S.%f")
        for item, n in hist.items():
            path, flags = item.rsplit(" ", 1)
            eprint(f"\033[36m{ts} \033[35m{n}x \033[33m{flags}\033[0m {path}\n")
            if self.do_stdout:
                print(f"{flags} {path}\n", end="")

    def _worker(self):
        cmd_pend = set()
        hist = OrderedDict()
        last_print = last_mod = time.time()
        while True:
            if cmd_pend and (not self.proc or self.proc.poll() is not None):
                self.proc = sp.Popen([self.cmd] + list(cmd_pend))
                cmd_pend = set()

            time.sleep(self.duration / 2.0)
            now = time.time()
            with self.mtx:
                if self.hist:
                    last_mod = now
                    hist.update(self.hist)
                    self.hist = OrderedDict()

            if not hist:
                continue

            if now - last_mod < self.duration and now - last_print < self.timeout:
                continue

            if self.cmd:
                # pfp = [x.rsplit(" ", 1) for x in hist.keys()]
                # pfp = [[x.split(","), y] for y, x in pfp]
                # pfp = ["".join(x[:1] for x in x) + "," + y for x, y in pfp]

                pfp = [x.rsplit(" ", 1)[0] for x in hist.keys()]
                cmd_pend.update(pfp)

            last_print = now
            self._print(hist)
            hist = OrderedDict()
            eprint("\n")


def rlist(top, ptn):
    yield top
    for parent, dirs, _ in os.walk(top):
        if ptn:
            dirs[:] = [d for d in dirs if not ptn.search(os.path.join(parent, d))]

        yield from [os.path.join(parent, d) for d in dirs]


def main():
    # fmt: off
    ap = argparse.ArgumentParser()
    ap.add_argument("-r", help="run command on each event batch")
    ap.add_argument("-x", help="exclude regex (full path)")
    ap.add_argument("-eo", action="store_true", help="filenames on stdout")
    ap.add_argument("-wd", metavar="SEC", type=float, default=0.2, help="debounce duration")
    ap.add_argument("-wt", metavar="SEC", type=float, default=2.0, help="debounce timeout")
    ap.add_argument("--df", help="enable periodic statvfs")
    ap.add_argument("PATH", help="top-level directory to monitor")
    ar = ap.parse_args()
    # fmt: on

    if ar.r and ar.wd == 0:
        eprint("FATAL: -wd must be non-zero for -r")
        sys.exit(1)

    ptn = None
    if ar.x:
        ptn = re.compile(ar.x)

    inotify = INotify()
    watch_flags = [INFlags.CREATE, INFlags.DELETE, INFlags.MODIFY, INFlags.DELETE_SELF]
    # watch_flags = INFlags.__members__.values()
    merged_flags = 0
    for flag in watch_flags:
        merged_flags |= flag

    wtab = {}
    for path in rlist(ar.PATH, ptn):
        try:
            wd = inotify.add_watch(path, merged_flags)
            wtab[wd] = path
            eprint(f"ok {path} \033[35m{wd}\033[0m\n")
        except Exception as ex:
            eprint(f"\033[1;31mNG {path}\n\033[0;33m └─{repr(ex)}\033[0m\n")

    if not wtab:
        eprint("\nFATAL: could not watch any directories\n")
        sys.exit(1)

    if ar.df:
        threading.Thread(target=diskmon, args=(ar.PATH,), daemon=True).start()

    sink = Sink(ar.wd, ar.wt, ar.r, ar.eo)
    while True:
        for ev in inotify.read():
            # eprint(f"\033[36m{time.time()}\033[0m {ev}\n")

            if ev.wd not in wtab:
                # IGNORED is emitted after DELETE_SELF
                continue

            path = os.path.join(wtab[ev.wd], ev.name)
            if ptn and ptn.search(path):
                continue

            flags = ev.mask
            sflags = ",".join([str(x)[6:] for x in INFlags.from_mask(flags)])
            del_self = flags & INFlags.DELETE_SELF
            if not del_self or ev.wd == 1:
                sink.put(path, sflags)

            if flags & INFlags.CREATE and flags & INFlags.ISDIR:
                try:
                    wd = inotify.add_watch(path, merged_flags)
                    wtab[wd] = path
                except OSError as ex:
                    eprint(f"E: {repr(ex)}\n")
                    continue

            elif del_self:
                del wtab[ev.wd]


if __name__ == "__main__":
    main()


# example:
"""
./rinotify.py /tmp/ 2>&1 | tee ~/rinotify.$(date +%Y-%m%d-%H%M%S)

cd /tmp/
mkdir q
mkdir q/a
echo h > q/s
echo h > q/a/s

rm -rf q

pv -L 2M /dev/zero > /tmp/nils
rm /tmp/nils
"""


# example: run script on event
"""
cat >~/rinotify-hello.sh <<'EOF'
#!/bin/bash
printf 'sh: [%s]\n' "$@"
printf 'sh: end\n'
EOF

~/rinotify.py . -r ~/rinotify-hello.sh
"""


# example: persistent rsync to ed@127.0.0.1:8022:~/foo
"""
cat >~/rinotify-rsync.sh <<'EOF'
#!/bin/bash
cfg="ssh -oControlPath=~/.ssh/irsync-%r@%h:%p -oControlMaster=auto -oControlPersist=60m -p8022"
rsync -e "$cfg" -SWaxHAX . ed@127.0.0.1:foo
EOF

cd ~/dev/stuff && ~/rinotify.py . -r ~/rinotify-rsync.sh
"""

